package ru.dbotthepony.kommons.arrays

import java.nio.ByteBuffer

abstract class Byte2DArray : Array2D() {
	abstract operator fun get(column: Int, row: Int): Byte
	abstract operator fun set(column: Int, row: Int, value: Byte)

	fun storeColumnRow(target: ByteBuffer) = matrixStoreColumnRow(this, target, Byte2DArray::get, ByteBuffer::put)
	fun storeRowColumn(target: ByteBuffer) = matrixStoreRowColumn(this, target, Byte2DArray::get, ByteBuffer::put)
	fun store(target: Byte2DArray) { target.load(this) }

	fun loadColumnRow(target: ByteBuffer) = matrixLoadColumnRow(this, target, Byte2DArray::set, ByteBuffer::get)
	fun loadRowColumn(target: ByteBuffer) = matrixLoadRowColumn(this, target, Byte2DArray::set, ByteBuffer::get)
	fun load(from: Byte2DArray) { matrixLoad(from, this, Byte2DArray::get, Byte2DArray::set) }
	fun storeTransposed(from: Byte2DArray) { matrixLoadTransposed(from, this, Byte2DArray::get, Byte2DArray::set) }

	open fun contentEquals(other: Byte2DArray): Boolean = matrixContentEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }

	override fun equals(other: Any?): Boolean {
		return matrixEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }
	}

	override fun hashCode(): Int {
		return matrixHashCode(this) { a, b, c -> a[b, c].hashCode() }
	}

	override fun toString(): String {
		return matrixToString(this, Byte2DArray::get)
	}

	val transposed: Byte2DArray get() {
		return allocate(rows, columns).also { it.load(this) }
	}

	private class Impl(
		override val columns: Int,
		override val rows: Int
	) : Byte2DArray() {
		private val mem = ByteArray(columns * rows)

		override fun get(column: Int, row: Int): Byte {
			try {
				return mem[column + row * columns]
			} catch (err: IndexOutOfBoundsException) {
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
			}
		}

		override fun set(column: Int, row: Int, value: Byte) {
			try {
				mem[column + row * columns] = value
			} catch (err: IndexOutOfBoundsException) {
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
			}
		}

		override fun equals(other: Any?): Boolean {
			if (other is Impl) return sizeEquals(other) && mem.contentEquals(other.mem)
			return super.equals(other)
		}

		override fun toString(): String {
			return "Byte2DArray" + super.toString()
		}
	}

	private class View(private val parent: Byte2DArray) : Byte2DArray() {
		override val rows: Int
			get() = parent.rows
		override val columns: Int
			get() = parent.columns

		override fun get(column: Int, row: Int): Byte {
			return parent[column, row]
		}

		override fun set(column: Int, row: Int, value: Byte) {
			throw UnsupportedOperationException("Read only view")
		}

		override fun equals(other: Any?): Boolean {
			return other is View && parent == other.parent
		}

		override fun hashCode(): Int {
			return parent.hashCode()
		}

		override fun toString(): String {
			return "View = $parent"
		}
	}

	companion object {
		@JvmStatic
		fun allocate(columns: Int, rows: Int): Byte2DArray {
			return Impl(columns, rows)
		}

		@JvmStatic
		fun unmodifiable(parent: Byte2DArray): Byte2DArray {
			if (parent is View)
				return parent
			else
				return View(parent)
		}

		@JvmField
		val ZERO: Byte2DArray = Impl(0, 0)
	}
}
