package ru.dbotthepony.kommons.io

import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import ru.dbotthepony.kommons.collect.ListenableMap
import ru.dbotthepony.kommons.collect.ListenableSet
import ru.dbotthepony.kommons.math.Decimal
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kommons.util.DelegateGetter
import ru.dbotthepony.kommons.util.DelegateSetter
import ru.dbotthepony.kommons.util.Listenable
import ru.dbotthepony.kommons.util.ListenableDelegate
import ru.dbotthepony.kommons.util.Observer
import java.io.DataInputStream
import java.io.DataOutputStream
import java.io.InputStream
import java.util.*
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Consumer
import java.util.function.Supplier
import kotlin.ConcurrentModificationException
import kotlin.collections.ArrayList
import kotlin.concurrent.withLock

/**
 * Synches attached delegates by serializing delta state on one end and deserializing on another.
 *
 * Precautions when using this class:
 * 1. Entire class is **not** thread safe, with the exception of attached delegates triggering [Listenable.addListener] listeners
 * (this includes calling [observe] from multiple threads, but why would you do this).
 * 2. [addListener] listeners are expected to be called once when [isDirty] turns true **on best effort**.
 * If syncher is running in highly congested environment (multiple threads are mutating attached delegates)
 * listeners may be called more than once IF [constructPayload] call is ongoing.
 * 3. [addListener] listeners on both [DelegateSyncher] and [DelegateSyncher.Slot]s are called on the same thread
 * where delegate mutation/observation/etc occurred. If you need to mutate [DelegateSyncher] internal structure (add/remove
 * slots) you **MUST** synchronize mutations by yourself if listener in question is expected to be called from multiple threads.
 * 4. When using [DelegateSyncher] in single-threaded context, all operations are legal at any point in time, including [DelegateSyncher.Slot]
 * addition/removal during [observe], [readPayload] or [addListener] callback on either [DelegateSyncher] or [DelegateSyncher.Slot].
 * 5. Calling [DelegateSyncher.AbstractSlot.remove] does not shift any subsequent slots down, leaving all slot IDs intact.
 * Gaps are filled when new slots are created.
 */
class DelegateSyncher : Listenable<Unit>, Observer {
	private val slots = ArrayList<AbstractSlot?>()
	private val observers = CopyOnWriteArrayList<AbstractSlot>()
	private val dirtySlots = ConcurrentLinkedQueue<AbstractSlot>()
	private val dirtySubs = Listenable.Void()
	private val gaps = IntAVLTreeSet()

	@Volatile
	private var changeset = 0

	var isRemote = false
		private set

	private val isDirtyF = AtomicBoolean()
	val isDirty: Boolean
		get() = isDirtyF.get()

	override fun observe(): Boolean {
		var any = false
		observers.forEach { any = it.observe() || any }
		return any
	}

	override fun addListener(listener: Runnable): Listenable.L {
		return dirtySubs.addListener(listener)
	}

	override fun addListener(listener: Consumer<Unit>): Listenable.L {
		return dirtySubs.addListener(listener)
	}

	fun constructPayload(): FastByteArrayOutputStream? {
		if (!isDirtyF.get()) return null

		val slots = ObjectAVLTreeSet<AbstractSlot>()

		do {
			var poll = dirtySlots.poll()

			while (poll != null) {
				slots.add(poll)
				poll = dirtySlots.poll()
			}
		} while (isDirtyF.compareAndSet(true, false))

		val stream = FastByteArrayOutputStream()
		val dataStream = DataOutputStream(stream)

		for (slot in slots) {
			dataStream.writeVarInt(slot.id + 1)
			slot.write(dataStream)
		}

		dataStream.writeVarInt(0)
		return stream
	}

	fun readPayload(stream: DataInputStream) {
		var nextId = stream.readVarInt()

		while (nextId > 0) {
			val slot = slots
			nextId = stream.readVarInt()
		}
	}

	fun readPayload(stream: InputStream) {
		readPayload(DataInputStream(stream))
	}

	abstract inner class AbstractSlot : Observer, Comparable<AbstractSlot> {
		protected val isDirtySlotF = AtomicBoolean(true)

		val isDirty: Boolean
			get() = isDirtySlotF.get()

		var isRemoved = false
			private set

		val id: Int

		init {
			val changeset = ++this@DelegateSyncher.changeset

			if (gaps.isNotEmpty()) {
				val gap = gaps.firstInt()
				gaps.remove(gap)
				id = gap
				check(slots[id] == null) { "Expected slot $id to be empty" }
				slots[id] = this
			} else {
				id = slots.size
				slots.add(this)
			}

			dirtySlots.add(this)

			if (changeset != ++this@DelegateSyncher.changeset) {
				throw ConcurrentModificationException("Manipulating DelegateSyncher across multiple threads")
			}

			if (isDirtyF.compareAndSet(false, true)) {
				dirtySubs.run()
			}
		}

		final override fun compareTo(other: AbstractSlot): Int {
			return id.compareTo(other.id)
		}

		protected abstract fun actuallyRemove()

		protected fun markDirty() {
			check(!isRemoved) { "This network slot was removed" }

			if (!isRemote && isDirtySlotF.compareAndSet(false, true)) {
				dirtySlots.add(this)

				if (!isDirtyF.get() && !dirtySlots.isEmpty() && isDirtyF.compareAndSet(false, true) && !dirtySlots.isEmpty()) {
					dirtySubs.run()
				}
			}
		}

		fun remove() {
			if (!isRemoved) {
				val changeset = ++this@DelegateSyncher.changeset

				isRemoved = true
				slots[id] = null
				observers.remove(this)

				if (id == slots.size - 1) {
					slots.removeAt(id)
				} else {
					gaps.add(id)
				}

				if (isDirtySlotF.get()) {
					dirtySlots.remove(this)
					this@DelegateSyncher.isDirtyF.set(dirtySlots.isNotEmpty())
				}

				actuallyRemove()

				if (changeset != ++this@DelegateSyncher.changeset) {
					throw ConcurrentModificationException("Manipulating DelegateSyncher across multiple threads")
				}
			}
		}

		abstract fun write(stream: DataOutputStream)
		abstract fun read(stream: DataInputStream)
	}

	inner class Slot<V>(private val parent: ListenableDelegate<V>, private val codec: StreamCodec<V>) : AbstractSlot(), ListenableDelegate<V> {
		private val subs = Listenable.Impl<V>()

		private val parentSub = parent.addListener(Consumer {
			markDirty()
			subs.accept(it)
		})

		override fun get(): V {
			check(!isRemoved) { "This network slot was removed" }
			return parent.get()
		}

		override fun accept(t: V) {
			check(!isRemoved) { "This network slot was removed" }
			parent.accept(t)
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved) { "This network slot was removed" }
			return subs.addListener(listener)
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved) { "This network slot was removed" }
			codec.write(stream, parent.get())
			isDirtySlotF.set(false)
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "This network slot was removed" }
			parent.accept(codec.read(stream))
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "This network slot was removed" }
			return false
		}

		override fun actuallyRemove() {
			subs.clear()
			parentSub.remove()
		}
	}

	inner class ObservedSlot<V>(private val parent: Delegate<V>, private val codec: StreamCodec<V>) : AbstractSlot(), ListenableDelegate<V> {
		private var lastValue: Any? = Mark
		private val localLock = ReentrantLock() // avoid deadlock when code inside subscribers cause another value observation
		private val subs = Listenable.Impl<V>()

		override fun actuallyRemove() {
			subs.clear()
			lastValue = Mark
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved) { "This network slot was removed" }
			codec.write(stream, parent.get())
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "This network slot was removed" }
			val read = codec.read(stream)
			parent.accept(read)

			localLock.withLock {
				lastValue = read
				subs.accept(read)
			}
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "This network slot was removed" }
			if (isRemote) return false

			val get = get()

			if (lastValue === Mark || !codec.compare(get, lastValue as V)) {
				localLock.withLock {
					if (lastValue === Mark || !codec.compare(get, lastValue as V)) {
						markDirty()
						lastValue = codec.copy(get)
						subs.accept(get)
						return true
					}
				}
			}

			return false
		}

		override fun get(): V {
			check(!isRemoved) { "This network slot was removed" }
			val get = get()

			if (!isRemote) {
				if (lastValue === Mark || !codec.compare(get, lastValue as V)) {
					localLock.withLock {
						if (lastValue === Mark || !codec.compare(get, lastValue as V)) {
							markDirty()
							lastValue = codec.copy(get)
							subs.accept(get)
						}
					}
				}
			}

			return get
		}

		override fun accept(t: V) {
			check(!isRemoved) { "This network slot was removed" }

			parent.accept(t)

			localLock.withLock {
				lastValue = codec.copy(t)
				subs.accept(t)
			}
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved) { "This network slot was removed" }
			return subs.addListener(listener)
		}
	}

	inner class ComputedSlot<V>(parent: Supplier<V>, private val codec: StreamCodec<V>) : AbstractSlot(), ListenableDelegate<V> {
		private val shadow = ListenableDelegate.CustomShadow(parent, codec::compare, codec::copy)

		override fun write(stream: DataOutputStream) {
			check(!isRemoved) { "This network slot was removed" }
			codec.write(stream, shadow.get())
			isDirtySlotF.set(false)
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "This network slot was removed" }
			shadow.accept(codec.read(stream))
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "This network slot was removed" }
			if (shadow.observe()) {
				markDirty()
				return true
			}

			return false
		}

		override fun get(): V {
			check(!isRemoved) { "This network slot was removed" }
			return shadow.get()
		}

		override fun accept(t: V) {
			check(!isRemoved) { "This network slot was removed" }
			shadow.accept(t)
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved) { "This network slot was removed" }
			return shadow.addListener(listener)
		}

		override fun actuallyRemove() {
			shadow.clearListeners()
		}
	}

	inner class SetSlot<E>(val value: ListenableSet<E>, private val codec: StreamCodec<E>) : AbstractSlot(), ListenableSet.SetListener<E>, Listenable<ListenableSet<E>> {
		private val sub = value.addListener(this)
		private val subs = Listenable.Impl<ListenableSet<E>>()
		private val backlog = ArrayList<DataOutputStream.() -> Unit>()

		override fun onClear() {
			check(!isRemoved) { "This network slot was removed" }
			synchronized(backlog) {
				markDirty()

				backlog.clear()
				backlog.add { write(CLEAR) }
			}
		}

		override fun onValueAdded(element: E) {
			check(!isRemoved) { "This network slot was removed" }
			val copy = codec.copy(element)

			synchronized(backlog) {
				markDirty()

				backlog.add {
					write(ADD)
					codec.write(this, copy)
				}
			}
		}

		override fun onValueRemoved(element: E) {
			check(!isRemoved) { "This network slot was removed" }
			val copy = codec.copy(element)

			synchronized(backlog) {
				markDirty()

				backlog.add {
					write(REMOVE)
					codec.write(this, copy)
				}
			}
		}

		override fun actuallyRemove() {
			backlog.clear()
			sub.remove()
			subs.clear()
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved) { "This network slot was removed" }
			synchronized(backlog) {
				isDirtySlotF.set(false)
				backlog.forEach { it(stream) }
				backlog.clear()
			}

			stream.write(END)
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "This network slot was removed" }
			var action = stream.read()

			while (true) {
				when (action) {
					ADD -> value.add(codec.read(stream))
					REMOVE -> value.remove(codec.read(stream))
					CLEAR -> value.clear()
					else -> break
				}

				action = stream.read()
			}

			subs.accept(value)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "This network slot was removed" }
			return false
		}

		override fun addListener(listener: Consumer<ListenableSet<E>>): Listenable.L {
			check(!isRemoved) { "This network slot was removed" }
			return subs.addListener(listener)
		}
	}

	inner class MapSlot<K, V>(val value: ListenableMap<K, V>, private val keyCodec: StreamCodec<K>, private val valueCodec: StreamCodec<V>) : AbstractSlot(), Listenable<ListenableMap<K, V>>, ListenableMap.MapListener<K, V> {
		private val sub = value.addListener(this)
		private val subs = Listenable.Impl<ListenableMap<K, V>>()
		private val backlog = ArrayList<DataOutputStream.() -> Unit>()

		override fun onClear() {
			check(!isRemoved) { "This network slot was removed" }
			synchronized(backlog) {
				markDirty()

				backlog.clear()
				backlog.add { write(CLEAR) }
			}
		}

		override fun onValueAdded(key: K, value: V) {
			check(!isRemoved) { "This network slot was removed" }
			val keyCopy = keyCodec.copy(key)
			val valueCopy = valueCodec.copy(value)

			synchronized(backlog) {
				markDirty()

				backlog.add {
					write(ADD)
					keyCodec.write(this, keyCopy)
					valueCodec.write(this, valueCopy)
				}
			}
		}

		override fun onValueRemoved(key: K, value: V) {
			check(!isRemoved) { "This network slot was removed" }
			val keyCopy = keyCodec.copy(key)

			synchronized(backlog) {
				markDirty()

				backlog.add {
					write(REMOVE)
					keyCodec.write(this, keyCopy)
				}
			}
		}

		override fun actuallyRemove() {
			backlog.clear()
			sub.remove()
			subs.clear()
		}

		override fun write(stream: DataOutputStream) {
			check(!isRemoved) { "This network slot was removed" }
			synchronized(backlog) {
				isDirtySlotF.set(false)
				backlog.forEach { it(stream) }
				backlog.clear()
			}

			stream.write(END)
		}

		override fun read(stream: DataInputStream) {
			check(!isRemoved) { "This network slot was removed" }
			var action = stream.read()

			while (true) {
				when (action) {
					ADD -> value.put(keyCodec.read(stream), valueCodec.read(stream))
					REMOVE -> value.remove(keyCodec.read(stream))
					CLEAR -> value.clear()
					else -> break
				}

				action = stream.read()
			}

			subs.accept(value)
		}

		override fun addListener(listener: Consumer<ListenableMap<K, V>>): Listenable.L {
			check(!isRemoved) { "This network slot was removed" }
			return subs.addListener(listener)
		}

		override fun observe(): Boolean {
			check(!isRemoved) { "This network slot was removed" }
			return false
		}
	}

	fun <V> add(delegate: ListenableDelegate<V>, codec: StreamCodec<V>): Slot<V> {
		return Slot(delegate, codec)
	}

	fun <V> add(delegate: Delegate<V>, codec: StreamCodec<V>): ObservedSlot<V> {
		return ObservedSlot(delegate, codec)
	}

	fun <V> add(delegate: Supplier<V>, codec: StreamCodec<V>): ComputedSlot<V> {
		return ComputedSlot(delegate, codec)
	}

	fun <E> add(delegate: ListenableSet<E>, codec: StreamCodec<E>): SetSlot<E> {
		return SetSlot(delegate, codec)
	}

	fun <K, V> add(delegate: ListenableMap<K, V>, keyCodec: StreamCodec<K>, valueCodec: StreamCodec<V>): MapSlot<K, V> {
		return MapSlot(delegate, keyCodec, valueCodec)
	}

	@JvmName("vbyte")
	@JvmOverloads
	fun byte(value: Byte = 0, getter: DelegateGetter<Byte> = DelegateGetter.passthrough(), setter: DelegateSetter<Byte> = DelegateSetter.passthrough()): Slot<Byte> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), ByteValueCodec)
	}

	@JvmName("vshort")
	@JvmOverloads
	fun short(value: Short = 0, getter: DelegateGetter<Short> = DelegateGetter.passthrough(), setter: DelegateSetter<Short> = DelegateSetter.passthrough()): Slot<Short> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), ShortValueCodec)
	}

	@JvmName("vchar")
	@JvmOverloads
	fun char(value: Char, getter: DelegateGetter<Char> = DelegateGetter.passthrough(), setter: DelegateSetter<Char> = DelegateSetter.passthrough()): Slot<Char> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), CharValueCodec)
	}

	@JvmName("vint")
	@JvmOverloads
	fun int(value: Int = 0, getter: DelegateGetter<Int> = DelegateGetter.passthrough(), setter: DelegateSetter<Int> = DelegateSetter.passthrough()): Slot<Int> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), VarIntValueCodec)
	}

	@JvmName("vlong")
	@JvmOverloads
	fun long(value: Long = 0L, getter: DelegateGetter<Long> = DelegateGetter.passthrough(), setter: DelegateSetter<Long> = DelegateSetter.passthrough()): Slot<Long> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), VarLongValueCodec)
	}

	@JvmName("vfloat")
	@JvmOverloads
	fun float(value: Float = 0f, getter: DelegateGetter<Float> = DelegateGetter.passthrough(), setter: DelegateSetter<Float> = DelegateSetter.passthrough()): Slot<Float> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), FloatValueCodec)
	}

	@JvmName("vdouble")
	@JvmOverloads
	fun double(value: Double = 0.0, getter: DelegateGetter<Double> = DelegateGetter.passthrough(), setter: DelegateSetter<Double> = DelegateSetter.passthrough()): Slot<Double> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), DoubleValueCodec)
	}

	@JvmName("vboolean")
	@JvmOverloads
	fun boolean(value: Boolean = false, getter: DelegateGetter<Boolean> = DelegateGetter.passthrough(), setter: DelegateSetter<Boolean> = DelegateSetter.passthrough()): Slot<Boolean> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), BooleanValueCodec)
	}

	@JvmOverloads
	fun string(value: String, getter: DelegateGetter<String> = DelegateGetter.passthrough(), setter: DelegateSetter<String> = DelegateSetter.passthrough()): Slot<String> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), BinaryStringCodec)
	}

	@JvmOverloads
	fun uuid(value: UUID, getter: DelegateGetter<UUID> = DelegateGetter.passthrough(), setter: DelegateSetter<UUID> = DelegateSetter.passthrough()): Slot<UUID> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), UUIDValueCodec)
	}

	@JvmOverloads
	fun decimal(value: Decimal = Decimal.ZERO, getter: DelegateGetter<Decimal> = DelegateGetter.passthrough(), setter: DelegateSetter<Decimal> = DelegateSetter.passthrough()): Slot<Decimal> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), DecimalValueCodec)
	}

	private object Mark

	/**
	 * Helper class which allows quick creation of [DelegateSyncher] from attached delegates.
	 */
	class Group {
		private val constructors = ArrayList<DelegateSyncher.() -> Unit>()

		fun <D : ListenableDelegate<V>, V> add(delegate: D, codec: StreamCodec<V>): D {
			constructors.add {
				add(delegate, codec)
			}

			return delegate
		}

		fun <D : Delegate<V>, V> add(delegate: D, codec: StreamCodec<V>): D {
			constructors.add {
				add(delegate, codec)
			}

			return delegate
		}

		fun <D : ListenableSet<E>, E> add(delegate: D, codec: StreamCodec<E>): D {
			constructors.add {
				add(delegate, codec)
			}

			return delegate
		}

		fun <D : ListenableMap<K, V>, K, V> add(delegate: D, keyCodec: StreamCodec<K>, valueCodec: StreamCodec<V>): D {
			constructors.add {
				add(delegate, keyCodec, valueCodec)
			}

			return delegate
		}

		@JvmName("vbyte")
		@JvmOverloads
		fun byte(value: Byte = 0, getter: DelegateGetter<Byte> = DelegateGetter.passthrough(), setter: DelegateSetter<Byte> = DelegateSetter.passthrough()): ListenableDelegate<Byte> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), ByteValueCodec)
		}

		@JvmName("vshort")
		@JvmOverloads
		fun short(value: Short = 0, getter: DelegateGetter<Short> = DelegateGetter.passthrough(), setter: DelegateSetter<Short> = DelegateSetter.passthrough()): ListenableDelegate<Short> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), ShortValueCodec)
		}

		@JvmName("vchar")
		@JvmOverloads
		fun char(value: Char, getter: DelegateGetter<Char> = DelegateGetter.passthrough(), setter: DelegateSetter<Char> = DelegateSetter.passthrough()): ListenableDelegate<Char> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), CharValueCodec)
		}

		@JvmName("vint")
		@JvmOverloads
		fun int(value: Int = 0, getter: DelegateGetter<Int> = DelegateGetter.passthrough(), setter: DelegateSetter<Int> = DelegateSetter.passthrough()): ListenableDelegate<Int> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), VarIntValueCodec)
		}

		@JvmName("vlong")
		@JvmOverloads
		fun long(value: Long = 0L, getter: DelegateGetter<Long> = DelegateGetter.passthrough(), setter: DelegateSetter<Long> = DelegateSetter.passthrough()): ListenableDelegate<Long> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), VarLongValueCodec)
		}

		@JvmName("vfloat")
		@JvmOverloads
		fun float(value: Float = 0f, getter: DelegateGetter<Float> = DelegateGetter.passthrough(), setter: DelegateSetter<Float> = DelegateSetter.passthrough()): ListenableDelegate<Float> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), FloatValueCodec)
		}

		@JvmName("vdouble")
		@JvmOverloads
		fun double(value: Double = 0.0, getter: DelegateGetter<Double> = DelegateGetter.passthrough(), setter: DelegateSetter<Double> = DelegateSetter.passthrough()): ListenableDelegate<Double> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), DoubleValueCodec)
		}

		@JvmName("vboolean")
		@JvmOverloads
		fun boolean(value: Boolean = false, getter: DelegateGetter<Boolean> = DelegateGetter.passthrough(), setter: DelegateSetter<Boolean> = DelegateSetter.passthrough()): ListenableDelegate<Boolean> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), BooleanValueCodec)
		}

		@JvmOverloads
		fun string(value: String, getter: DelegateGetter<String> = DelegateGetter.passthrough(), setter: DelegateSetter<String> = DelegateSetter.passthrough()): ListenableDelegate<String> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), BinaryStringCodec)
		}

		@JvmOverloads
		fun uuid(value: UUID, getter: DelegateGetter<UUID> = DelegateGetter.passthrough(), setter: DelegateSetter<UUID> = DelegateSetter.passthrough()): ListenableDelegate<UUID> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), UUIDValueCodec)
		}

		@JvmOverloads
		fun decimal(value: Decimal = Decimal.ZERO, getter: DelegateGetter<Decimal> = DelegateGetter.passthrough(), setter: DelegateSetter<Decimal> = DelegateSetter.passthrough()): ListenableDelegate<Decimal> {
			return add(ListenableDelegate.maskSmart(value, getter, setter), DecimalValueCodec)
		}

		fun <E> set(codec: StreamCodec<E>, backing: MutableSet<E> = ObjectOpenHashSet()): ListenableSet<E> {
			val delegate = ListenableSet(backing)
			constructors.add { add(delegate, codec) }
			return delegate
		}

		fun <K, V> map(keyCodec: StreamCodec<K>, valueCodec: StreamCodec<V>, backing: MutableMap<K, V> = Object2ObjectOpenHashMap()): ListenableMap<K, V> {
			val delegate = ListenableMap(backing)
			constructors.add { add(delegate, keyCodec, valueCodec) }
			return delegate
		}

		fun add(other: Group): Group {
			constructors.addAll(other.constructors)
			return this
		}

		fun create(): DelegateSyncher {
			return addInto(DelegateSyncher())
		}

		fun addInto(syncher: DelegateSyncher): DelegateSyncher {
			constructors.forEach { it(syncher) }
			return syncher
		}
	}

	companion object {
		const val END = 0
		const val CLEAR = 1
		const val ADD = 2
		const val REMOVE = 3
	}
}
